"""
    Copyright (C) 2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import gzip
import xml.etree.ElementTree as ET
from io import BytesIO

import requests

from .Addon import Addon


class Repository():
    def __init__(self, version, path):
        super(Repository, self).__init__()
        self.version = version
        self.path = path

        # Recover from unreliable mirrors
        session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=5)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        content = session.get(path, timeout=(30, 30)).content

        if path.endswith('.gz'):
            with gzip.open(BytesIO(content), 'rb') as xml_file:
                content = xml_file.read()

        tree = ET.fromstring(content)
        self.addons = []
        for addon in tree.findall("addon"):
            self.addons.append(Addon(addon))

    def __contains__(self, addonId):
        for addon in self.addons:
            if addon.id == addonId:
                return True
        return False

    def find(self, addonId):
        for addon in self.addons:
            if addon.id == addonId:
                return addon
        return None

    def rdepends(self, addonId):
        rdepends = []
        for addon in self.addons:
            if addon.dependsOn(addonId):
                rdepends.append(addon)
        return rdepends



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import copy
import logging
from os.path import dirname, exists, join

import xmlschema

from . import ValidKodiVersions
from .record import INFORMATION, PROBLEM, Record
from .report import Report

XML_SCHEMA = join(dirname(__file__), 'xml_schema')
LOGGER = logging.getLogger(__name__)


def schemas(report: Report, CONST_parsed_xml, branch_name):
    """validates XML file with existing schemas
    :parsed_xml: parsed data from an XML file
    """
    parsed_xml = copy.copy(CONST_parsed_xml)
    failed, metadatacount, valid = _validation_checks(report, parsed_xml, branch_name)

    root_schema = join(XML_SCHEMA, 'addon.xsd')
    if not _validate(parsed_xml, root_schema):
        report.add(Record(PROBLEM, "Schema validation failed for root addon element"))
        valid = False

    if metadatacount == 0:
        report.add(Record(PROBLEM, 'Metadata missing/occurred more than once'))

    if valid:
        report.add(Record(INFORMATION, "Valid XML file found"))
    elif failed:
        report.add(Record(PROBLEM, "Schema validation failed for the following points: %s " % ''.join(failed)))


def _validation_checks(report: Report, parsed_xml, branch_name):
    """Perform certain checks on the XML elements and extension point
    """
    metadatacount = 0
    valid = True
    failed = []
    valid_points = {
        'kodi.context.item': 'contextitem.xsd',
        'kodi.game.controller': 'controller.xsd',
        'kodi.resource.games': 'games.xsd',
        'kodi.resource.images': 'images.xsd',
        'kodi.resource.language': 'language.xsd',
        'kodi.addon.metadata': 'metadata.xsd',
        'kodi.resource.uisounds': 'uisounds.xsd',
        'xbmc.addon.metadata': 'metadata.xsd',
        'xbmc.addon.repository': 'repository.xsd',
        'xbmc.gui.skin': 'skin.xsd',
        'xbmc.metadata.scraper.albums': 'scraper.xsd',
        'xbmc.metadata.scraper.artists': 'scraper.xsd',
        'xbmc.metadata.scraper.movies': 'scraper.xsd',
        'xbmc.metadata.scraper.musicvideos': 'scraper.xsd',
        'xbmc.metadata.scraper.tvshows': 'scraper.xsd',
        'xbmc.metadata.scraper.library': 'scraper.xsd',
        'xbmc.python.script': 'pythonscript.xsd',
        'xbmc.python.lyrics': 'script.xsd',
        'xbmc.python.weather': 'script.xsd',
        'xbmc.python.library': 'script.xsd',
        'xbmc.python.pluginsource': 'pluginsource.xsd',
        'xbmc.python.module': 'script.xsd',
        'xbmc.service': 'service.xsd',
        'xbmc.subtitle.module': 'script.xsd',
        'xbmc.ui.screensaver': 'script.xsd',
        'xbmc.webinterface': 'webinterface.xsd',
    }

    for extension in parsed_xml.findall("extension"):
        extension_point = extension.get("point")
        if extension_point in valid_points:
            schema_path = check_version(branch_name, valid_points[extension_point])

            if schema_path:
                if extension_point in ("xbmc.addon.metadata", "kodi.addon.metadata"):
                    metadatacount += 1

                if not exists(schema_path):
                    report.add(Record(PROBLEM, "%s doesn't exist" % schema_path))
                    continue

                if not _validate(extension, schema_path):
                    failed.append(extension_point)
                    valid = False

                parsed_xml.remove(extension)
            else:
                report.add(Record(PROBLEM, "schema for %s doesn't exists" % extension_point))
        else:
            report.add(Record(PROBLEM, "%s is not a valid extension point" % extension_point))
            valid = False

    return failed, metadatacount, valid


def _validate(xml, schemapath):
    """Validates given XML file/element with a given schema

    :xml: parsed xml element/file
    :schemapath: path to a schema file
    """
    schema = xmlschema.XMLSchema(schemapath)
    for error in schema.iter_errors(xml):
        print(error)
    return schema.is_valid(xml)


def check_version(branch_name, schema_file):
    all_branches = ValidKodiVersions[::-1]
    branches = all_branches[all_branches.index(branch_name)::1]
    for branch in branches:
        file = branch + '_' + schema_file
        schema_file_path = join(XML_SCHEMA, file)
        if exists(schema_file_path):
            return schema_file_path
    return None



"""
    Copyright (C) 2018-201 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

PROBLEM = "ERROR"
WARNING = "WARN"
INFORMATION = "INFO"


class Record():
    def __init__(self, log_level, message):
        """
        Create a record which is the low level entry in Report to represent logs.
        :param log_level: "ERROR", "WARN" or "INFO"
        :param message: the actual log message
        """
        self.log_level = log_level
        self.message = message

    def add(self, record):
        pass

    def __str__(self):
        """
        Text representation of record.
        :return: text representation of report
        """
        return "%s: %s" % (self.log_level, self.message)



"""
    Copyright (C) 2019 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

from packaging.version import parse, Version, LegacyVersion
from kodi_addon_checker import ValidKodiVersions


class AddonVersion():
    def __init__(self, version):
        self.version = parse(str(version).lower())
        # non PEP440 compliant versions (legacy), for beta and alpha versions
        # convert them into PEP440 format: 1.1.0~beta01 -> 1.1.0beta01
        if self._islegacy_devversion:
            self.version = Version(version.replace("~", ""))

    @property
    def _islegacy_devversion(self):
        return isinstance(self.version, LegacyVersion) and \
            ("~beta" or "~alpha" in self.version)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return self.version < other.version

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return self.version <= other.version

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.version == other.version

    def __ne__(self, other):
        return not isinstance(other, self.__class__) or self.version != other.version

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return self.version > other.version

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return self.version >= other.version

    def __repr__(self):
        return str(self.version)


class KodiVersion():
    def __init__(self, version: str):
        if version not in ValidKodiVersions:
            raise ValueError("Invalid KodiVersion")
        super(KodiVersion, self).__init__()
        self.version = version

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return ValidKodiVersions.index(self.version) < ValidKodiVersions.index(other.version)

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return ValidKodiVersions.index(self.version) <= ValidKodiVersions.index(other.version)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.version == other.version

    def __ne__(self, other):
        return not isinstance(other, self.__class__) or self.version != other.version

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return ValidKodiVersions.index(self.version) > ValidKodiVersions.index(other.version)

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        return ValidKodiVersions.index(self.version) >= ValidKodiVersions.index(other.version)

    def __repr__(self):
        return str(self.version)



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import logging
import os
import xml.etree.ElementTree as ET

from .check_dependencies import VERSION_ATTRB
from .record import INFORMATION, PROBLEM, Record, WARNING
from .report import Report
from .versions import AddonVersion, KodiVersion


LOGGER = logging.getLogger(__name__)


def check_for_existing_addon(report: Report, addon_path: str, all_repo_addons: dict, pr: bool,
                             kodi_version: KodiVersion):
    """Check if addon submitted already exists or not
        :addon_path: path of the addon
        :all_repo_addons: dictionary return by all_repo_addon() function
    """

    addon_xml = os.path.join(addon_path, "addon.xml")
    addon_name, addon_version = _get_addon_name(addon_xml)
    addon_details = {'name': addon_name, 'version': addon_version}

    is_new_addon = True
    for branch, repo in sorted(all_repo_addons.items(), reverse=True):

        # Addon submission must be higher than the versions already available in lower
        # branches to allow users to receive the update (especially if run with --pr)
        if KodiVersion(branch) <= kodi_version and addon_name in repo:
            is_new_addon = False
            _check_version_higher(report, addon_details, branch, repo.find(addon_name).version, pr)

        # Addon submission must be lower than the versions already available in upper repo branches
        # if that branch corresponds to a breaking change (e.g. version of addon in matrix version >
        # version of addon in gotham) since xbmc.python is not backwards compatible.
        elif KodiVersion(branch) > kodi_version and addon_name in repo and \
            not _is_pythonabi_compatible(repr(kodi_version), branch):
            is_new_addon = False
            _check_version_lower(report, addon_details, branch, repo.find(addon_name).version, pr)

    if is_new_addon:
        report.add(Record(INFORMATION, "This is a new addon"))


def _get_addon_name(xml_path: str):
    """returns name and version of the addon
        :xml_path: path of the xml file
    """

    tree = ET.parse(xml_path).getroot()
    return (tree.get("id"), tree.get("version"))


def _is_pythonabi_compatible(target_branch, upper_branch):
    """returns true if the target_branch for this addon is backwards compatible with a
        given upper branch. E.g. if kodi isengard python abi is compatible with leia python abi.
        :target_branch: the branch the addon lives in (or is being PR'd to)
        : upper_branch: an upper branch that also contains an addon with the same addon id
    """
    return VERSION_ATTRB['xbmc.python'][upper_branch]["min_compatible"] == \
        VERSION_ATTRB['xbmc.python'][target_branch]["min_compatible"]


def _check_version_higher(report: Report, addon_details, branch, repo_addons_version, pr):
    """Check the version in lower branch is lower than the addon version being submitted
       thus allowing for addon updates.

        :addon_details:       a dict containing name and version of the addon {'name': .., 'version': ..}
        :branch:              branch of the addon present in Kodi repository
        :repo_addons_version: version of addon present in Kodi repository
        :pr:                  boolean value indicating whether the check is
                              running on pull request or not
    """
    addon_name = addon_details.get('name')
    addon_version = addon_details.get('version')

    if AddonVersion(addon_version) < AddonVersion(repo_addons_version) or \
        (AddonVersion(addon_version) == AddonVersion(repo_addons_version) and pr):
        report.add(
            Record(
                PROBLEM,
                "%s addon already exists with a higher or equal version: %s in %s branch. Users in %s won't " \
                "be able to receive the addon update."
                % (addon_name, repo_addons_version, branch, branch)
            )
        )
    else:
        if pr:
            report.add(Record(INFORMATION, "%s addon also exists in %s branch but with update compatible version: %s"
                              % (addon_name, branch, repo_addons_version)))


def _check_version_lower(report: Report, addon_details, branch, repo_addons_version, pr):
    """Check the version in upper branch is higher than the addon version being submitted
       thus allowing for kodi migration and inherent addon update.

        :addon_details:       a dict containing name and version of the addon {'name': .., 'version': ..}
        :branch:              branch of the addon present in Kodi repository
        :repo_addons_version: version of addon present in Kodi repository
        :pr:                  boolean value indicating whether the check is
                              running on pull request or not
    """
    addon_name = addon_details.get('name')
    addon_version = addon_details.get('version')

    if AddonVersion(addon_version) > AddonVersion(repo_addons_version) or \
        (AddonVersion(addon_version) == AddonVersion(repo_addons_version) and pr):
        report.add(
            Record(
                PROBLEM if pr else WARNING,
                "%s addon already exists with a lower or equal version: %s in %s branch. Users migrating " \
                "to kodi version %s won't be able to receive the addon update"
                % (addon_name, repo_addons_version, branch, branch)
            )
        )
    else:
        if pr:
            report.add(Record(INFORMATION, "%s addon also exists in %s branch but with migration compatible version: %s"
                              % (addon_name, branch, repo_addons_version)))



"""
    Copyright (C) 2017-2018 Team Kodi
    This file is part of Kodi - kodi.tv

    SPDX-License-Identifier: GPL-3.0-only
    See LICENSES/README.md for more information.
"""

import os
import re

import polib

from . import handle_files
from .common import relative_path
from .record import INFORMATION, PROBLEM, WARNING, Record
from .report import Report


def check_for_legacy_strings_xml(report: Report, addon_path: str):
    """Find for the string.xml file in addon which was used in old versions
        :addon_path: path of the addon
    """
    for file in handle_files.find_files_recursive("strings.xml", os.path.join(addon_path, "resources", "language")):
        report.add(
            Record(PROBLEM, "Found %s please migrate to strings.po." % relative_path(file)))


def find_blacklisted_strings(report: Report, addon_path: str, problems: list, warnings: list, file_types: list):
    """Find for any blacklisted strings in the addons files
        :addon_path: Path of theh addon
        :problems: List of all the strings that will cause problem being in an addon
        :warnings: List of all the strings that shouldn't be in addon
                        but doesn't cause any problem
        :file_type: List of the whitelisted files to look into
    """
    for result in handle_files.find_in_file(addon_path, problems, file_types):
        report.add(Record(PROBLEM, "Found blacklisted term %s in file %s:%s (%s)"
                          % (result["term"], result["searchfile"], result["linenumber"], result["line"])))

    for result in handle_files.find_in_file(addon_path, warnings, file_types):
        report.add(Record(WARNING, "Found blacklisted term %s in file %s:%s (%s)"
                          % (result["term"], result["searchfile"], result["linenumber"], result["line"])))


def check_for_invalid_strings_po(report: Report, file_index: list):
    """Validate strings.po files
        :file_index: list having names and path of all the files present in addon
    """
    po_file_index = [f for f in file_index if f["name"] == "strings.po"]
    report_made = False

    for po_file in po_file_index:
        full_path = os.path.join(po_file["path"], po_file["name"])

        with open(full_path, "r", encoding="utf-8") as f:
            try:
                contents = f.read()
            except UnicodeDecodeError:
                report_made = True
                report.add(Record(PROBLEM, "Invalid PO file %s: File is not saved with UTF-8 encoding"
                                  % (relative_path(full_path))))
                continue

        if not contents:
            report_made = True
            report.add(Record(PROBLEM, "Invalid PO file %s: File is empty" % (relative_path(full_path))))
            continue

        if "\r\n" in contents:
            report_made = True
            report.add(Record(WARNING, "Windows line endings found in %s, consider converting to Linux line endings."
                              % relative_path(full_path)))

        header = contents[:contents.find("msgctxt \"#")]
        if not re.search(r'msgid ""\s+msgstr ""', header):
            # This is only required by polib if metadata follows, Kodi requires this regardless of metadata
            report_made = True
            report.add(Record(PROBLEM, "Invalid PO file %s:\nMissing required header:\n"
                                       "\tmsgid \"\"\n\tmsgstr \"\"" % (relative_path(full_path))))

        if contents[0] == "\ufeff":
            report_made = True
            report.add(Record(PROBLEM, "Invalid PO file %s: File contains BOM (byte order mark)"
                              % (relative_path(full_path))))
            continue

        try:
            polib.pofile(full_path, encoding="utf-8")
        except OSError as error:
            # raised on the first syntax error
            message = str(error)
            patterns = [r"\(line (?P<line_num>[0-9]+)\)\s*:\s*(?P<message>[\s\S]+)\s*$",
                        r"(?P<message>Syntax error)[\s\S]+\(line (?P<line_num>[0-9]+)\)\s*$"]
            for pattern in patterns:
                match = re.search(pattern, message)
                if match:
                    # restructure message to remove file and path
                    message = "%s on line %s" % (match.group("message"), match.group("line_num"))
                    break

            report_made = True
            report.add(Record(PROBLEM, "Invalid PO file %s: %s" % (relative_path(full_path), message)))

    if po_file_index and not report_made:
        report.add(Record(INFORMATION, "PO files are valid"))
